import unittest
import xml.etree.ElementTree as ET

from bidon.util.xml import transform


class UtilXMLTestCase(unittest.TestCase):
  def test_transform(self):
    tree = ET.parse("tests/fixtures/purchase-order.xml")
    root = tree.getroot()
    result = transform(root, dict(
      no="PurchaseOrderNumber",
      dt="OrderDate",
      to=("Address[@Type='Shipping']", dict(
        name=("Name", ),
        country=("Country", ))),
      items=("Items/Item", dict(no="PartNumber", name=("ProductName", )), "*")))

    self.assertEqual(result, dict(
      no="99503",
      dt="1999-10-20",
      to=dict(name="Ellen Adams", country="USA"),
      items=[
        dict(no="872-AA", name="Lawnmower"),
        dict(no="926-AA", name="Baby Monitor")
      ]))



import os
from collections import namedtuple

from bidon.data import DataAccess, ModelAccess


DA_CORE = None
TEST_ROWCOUNT = False
TEST_CALLPROC = False


def configure(database="postgres"):
  if database == "postgres":
    configure_postgres(None)
  elif database == "mysql":
    configure_mysql(None)
  elif database == "sqlite":
    configure_sqlite(None)


def configure_postgres(args):
  from psycopg2.extras import Json
  from psycopg2.extensions import register_adapter

  from bidon.data.data_access_core import get_pg_core

  global DA_CORE, TEST_ROWCOUNT, TEST_CALLPROC

  DA_CORE = get_pg_core("dbname=bidon_test user=postgres host=localhost")
  TEST_ROWCOUNT = True
  TEST_CALLPROC = True

  register_adapter(dict, lambda d: Json(d))


def configure_mysql(args):
  from pymysql.cursors import Cursor

  from bidon.data.data_access_core import get_mysql_core

  global DA_CORE, TEST_ROWCOUNT, TEST_CALLPROC

  class NamedTupleCursorMixin(object):
    def _do_get_result(self):
      super(NamedTupleCursorMixin, self)._do_get_result()
      if self.description:
        fields = [f.name for f in self._result.fields]
        Row = namedtuple("Row", fields)
        self._rows = [Row(*row) for row in self._rows]

  class NamedTupleCursor(NamedTupleCursorMixin, Cursor):
    pass

  defaults_file = os.path.join(os.path.expanduser("~"), ".mysql-defaults/localhost")
  DA_CORE = get_mysql_core(
    dict(read_default_file=defaults_file, database="bidon_test"),
    cursor_factory=NamedTupleCursor)
  TEST_ROWCOUNT = True
  TEST_CALLPROC = False


def configure_sqlite(args):
  from bidon.data.data_access_core import get_sqlite_core

  global DA_CORE, TEST_ROWCOUNT, TEST_CALLPROC

  def get_namedtuple_factory():
    fields = []
    Row = None
    def namedtuple_factory(cursor, row):
      nonlocal fields, Row
      _fields = [col[0] for col in cursor.description]
      if _fields != fields:
        fields = _fields
        Row = namedtuple("Row", fields)
      return Row(*row)
    return namedtuple_factory

  DA_CORE = get_sqlite_core("tests/fixtures/test.sqlite3", cursor_factory=get_namedtuple_factory())
  TEST_ROWCOUNT = False
  TEST_CALLPROC = False


def get_data_access():
  return DataAccess(DA_CORE)


def get_model_access():
  return ModelAccess(DA_CORE)


from .data import *
from .util import *
from .spreadsheet import *
from .test_data_table import DataTableTestCase
from .test_json_patch import JsonPatchTestCase



"""Contains the ForeignModelWrapper class."""
from . import ModelBase
from bidon.util import xml


class ForeignModelWrapper(ModelBase):
  """Provides a simple way to map incoming data into a model base."""
  timestamps = None
  table_name = None
  primary_key_name = None
  primary_key_is_auth = False
  mapping_format = "raw"
  mapping = dict()
  submodels = dict()

  @classmethod
  def apply_mapping(cls, from_dict, *, mapping=None):
    """Apply the class' mappings to a dictionary to create a new dictionary that can be used to
    initialize a new class instance.
    """
    if mapping is None:
      mapping = cls.mapping

    result = dict()
    for to_key, (from_key, transform) in mapping.items():
      value = from_dict.get(from_key)
      if value is not None and transform is not None:
        try:
          value = transform(value)
        except Exception as ex:
          raise Exception("Error with {0}->{1} conversion value {2}.\nAll data: {3}".format(
            from_key, to_key, value, from_dict)) from ex
      result[to_key] = value
    return result

  @classmethod
  def create(cls, from_item, *, mapping_format=None, mapping=None):
    """Passes from_item to apply_mapping and uses the result to construct a new class instance."""
    if mapping_format is None:
      mapping_format = cls.mapping_format

    if mapping_format == "raw":
      model = cls(cls.apply_mapping(from_item, mapping=mapping))
    elif mapping_format == "xml":
      model = cls(xml.transform(from_item, mapping or cls.mapping))
    else:
      raise ValueError("mapping_format must be one of {'raw', 'xml'}")

    for key, submodel in cls.submodels.items():
      value = getattr(model, key)

      if isinstance(value, list):
        new_value = [submodel(lv) for lv in value]
      else:
        new_value = submodel(value)

      setattr(model, key, new_value)

    return model

  @classmethod
  def map(cls, iterables, *, mapping_format=None, mapping=None):
    for idx, iterable in enumerate(iterables):
      try:
        yield cls.create(iterable, mapping_format=mapping_format, mapping=mapping)
      except Exception as ex:
        raise Exception("An error occurred with item {0}".format(idx)) from ex



import unittest
from datetime import date, time, datetime
from decimal import Decimal

from bidon.spreadsheet import Cell
from bidon.spreadsheet.csv import CSVWorkbook
from bidon.data_table import DataTable


def get_cell_value(cell):
    return cell.value


def set_cell_value(row, index, value):
    row[index].value = value


def is_cell_empty(cell):
    return cell.value is None


class DataTableTestCase(unittest.TestCase):
  def get_data(self):
    return CSVWorkbook("tests/fixtures/data-table.csv").sheets(0).to_cell_table()

  def get_data_table(self, data=None):
    return DataTable(data or self.get_data(), get_cell_value, set_cell_value, is_cell_empty)

  def test_index(self):
    data_table = self.get_data_table()
    self.assertEqual(data_table[(1, 0)].value, "Saturday Kids 5")

  def test_initialize(self):
    data_table = self.get_data_table()
    self.assertEqual(11, len(data_table.rows))
    self.assertEqual(11, data_table.nrows)
    self.assertEqual(4, data_table.ncols)

  def test_serialize(self):
    data_table = self.get_data_table()
    data = data_table.serialize(lambda c: c.value)
    self.assertEqual(data[0][0], "Name")
    self.assertEqual(data[1][1], "2014-09-20")

  def test_rows_to_dicts(self):
    data_table = self.get_data_table()
    dicts = list(data_table.rows_to_dicts(lambda c: c.value))
    self.assertEqual(dicts[0], dict(Name="Saturday Kids 5", Date="2014-09-20", Time="5:00 PM", Attendance=55))

  def test_trim_empty_rows(self):
    data = self.get_data()
    data.append([Cell() for i in data[0]])
    data_table = self.get_data_table(data)
    self.assertEqual(12, data_table.nrows)
    data_table.trim_empty_rows()
    self.assertEqual(11, data_table.nrows)
    self.assertEqual(11, len(data_table.rows))
    data.insert(0, [Cell() for i in data[0]])
    data.insert(5, [Cell() for i in data[0]])
    data_table = self.get_data_table(data)
    self.assertEqual(14, data_table.nrows)
    data_table.trim_empty_rows()
    self.assertEqual(13, data_table.nrows)
    self.assertEqual(13, len(data_table.rows))

  def test_trim_empty_columns(self):
    data = self.get_data()
    for row in data:
      row.extend([Cell(), Cell()])
      row.insert(0, Cell())
      row.insert(2, Cell())
    data_table = self.get_data_table(data)
    self.assertEqual(8, data_table.ncols)
    data_table.trim_empty_columns()
    self.assertEqual(6, data_table.ncols)
    self.assertEqual(6, len(data_table.rows[0]))

  def test_clean_values(self):
    data = [[" a  b  ", 3.898989, None], [None, None, None]]
    data_table = DataTable(data)
    data_table.clean_values()
    self.assertEqual(data_table.rows[0][0], "a b")
    self.assertEqual(data_table.rows[0][1], Decimal("3.89"))
    data_table.trim_empty_rows()
    data_table.trim_empty_columns()
    self.assertEqual(data_table.serialize(), [["a b", Decimal("3.89")]])
    data_table = DataTable(data)
    data_table.cleanup()
    self.assertEqual(data_table.serialize(), [["a b", Decimal("3.89")]])

  def test_empty_table(self):
    data = []
    data_table = DataTable(data)
    self.assertEqual(data_table.nrows, 0)
    self.assertEqual(data_table.ncols, 0)
    data_table.cleanup()
    data_table.serialize()



"""Methods for transforming xml.ElementTree nodes."""


def get_text(element):
  """Returns the text of element."""
  return element.text


def get_attribute(element, name):
  """Returns an attribute from an element."""
  return element.attrib[name]


def to_dict(element, transform_def):
  """Transform an element into a dict.

  transform_def is a dict. The return value's keys will match transform_def, and the values will
  be the result from transform(element, transform_def[key]).
  """
  return {k: transform(element, a) for k, a in transform_def.items()}


def transform(element, method=None):
  """A flexible element transformer.

  method can be one of 5 types: None, str, dict, tuple, callable. The behaviors are as follows:

  None - return the text of the element
  str - return the attribute of the element whose name matches the argument
  dict - transform the element into a dict, whose keys are the keys of the given dict, and whose
         values are the result from passing the element and the parameter dict's value to this
         function.
  tuple - call transform on one or more children. The tuple can be either a 1-tuple, 2-tuple or
          3-tuple. They are translated as follows:

          (xpath[, method[, index]])

          xpath - the xpath expression to get the children from the element
          method - any value that can be passed as this function's `method` argumet. Defaults to
                   None
          index - either the string '*' or a child index or None. Defaults to 0. If index is '*',
                  then the result will be a list using all the children found by xpath. If the index
                  is None, then xpath is interpreted instead as an attribute reference. This allows
                  you to pass in callables for the method to be called on attributes.
  callable - returns the result of method(element)
  """
  # pylint: disable=too-many-return-statements

  if method is None:
    return get_text(element)

  elif isinstance(method, str):
    return get_attribute(element, method)

  elif isinstance(method, dict):
    return to_dict(element, method)

  elif isinstance(method, tuple) and len(method) in (1, 2, 3):
    if len(method) == 1:
      xpath = method[0]
      method_ = None
      index = 0

    elif len(method) == 2:
      xpath, method_ = method
      index = 0

    else:
      xpath, method_, index = method

    try:
      if index == "*":
        return list(transform_all(element.findall(xpath), method_))
      elif index is None:
        raw_val = get_attribute(element, xpath)
        return raw_val if method_ is None else method_(raw_val)
      else:
        return transform(element.findall(xpath)[index], method_)
    except Exception as ex:
      raise Exception("An error occurred with path {}".format(xpath)) from ex

  elif callable(method):
    return method(element)

  else:
    raise ValueError("Unrecognized transformation method")


def transform_all(elements, method):
  """Generates a sequence of transformed elements."""
  for element in elements:
    yield transform(element, method)



"""Contains an implementation of WorkbookBase and WorksheetBase for working with CSV documents."""
import csv

from .base import CellMode, WorksheetBase, WorkbookBase


class CSVWorksheet(WorksheetBase):
  """Specialization of WorksheetBase for working with CSV documents."""
  def __init__(self, raw_sheet, ordinal):
    """Initialize the CSVWorksheet instance."""
    super().__init__(raw_sheet, ordinal)
    self.name = "Sheet 1"
    self.nrows = len(self.raw_sheet)
    self.ncols = max([len(r) for r in self.raw_sheet])

  def parse_cell(self, cell, coords, cell_mode=CellMode.cooked):
    """Tries to convert the value first to an int, then a float and if neither is
    successful it returns the string value.
    """
    try:
      return int(cell)
    except ValueError:
      pass
    try:
      return float(cell)
    except ValueError:
      pass
    # TODO Check for dates?
    return cell

  def get_row(self, row_index):
    """Returns the row at row_index."""
    return self.raw_sheet[row_index]


class CSVWorkbook(WorkbookBase):
  """Implementation of WorkbookBase for working with CSV documents."""
  def iterate_sheets(self, *args, **kwargs):
    """Opens self.filename and reads it with a csv reader."""
    with open(self.filename, "r") as rfile:
      reader = csv.reader(rfile, *args, **kwargs)
      yield list(reader)

  def get_worksheet(self, raw_sheet, index):
    """Creates a CSVWorksheet instance from the raw sheet."""
    return CSVWorksheet(raw_sheet, index)



from datetime import date, time, datetime
from enum import Enum


class CellMode(Enum):
  """Enum of cell modes."""
  raw = 0x01
  cooked = 0x02


class Cell(object):
  """Represents a cell in a spreadsheet."""
  def __init__(self, value=None, note=None):
    """Initialize the cell instance."""
    self.value = value
    self.note = note

  def copy(self):
    """Copy the cell object."""
    return Cell(self.value, self.note)

  def __str__(self):
    """Returns a string representation of a cell object."""
    return "Cell(value={0}, note={1})".format(self.value, self.note)


class WorksheetBase(object):
  """The base class for all worksheets. For each implemented spreadsheet type, it is necessary to
  write a subclass of this class that implements the necessary methods."""
  def __init__(self, raw_sheet, ordinal):
    """Initialize the WorksheetBase instance."""
    self.raw_sheet = raw_sheet
    self.ordinal = ordinal
    self.name = None
    self.nrows = -1
    self.ncols = -1

  def to_cell_table(self, merged=True):
    """Returns a list of lists of Cells with the cooked value and note for each cell."""
    new_rows = []
    for row_index, row in enumerate(self.rows(CellMode.cooked)):
      new_row = []
      for col_index, cell_value in enumerate(row):
        new_row.append(Cell(cell_value, self.get_note((col_index, row_index))))
      new_rows.append(new_row)
    if merged:
      for cell_low, cell_high in self.merged_cell_ranges():
        anchor_cell = new_rows[cell_low[1]][cell_low[0]]
        for row_index in range(cell_low[1], cell_high[1]):
          for col_index in range(cell_low[0], cell_high[0]):
            # NOTE: xlrd occassionally returns ranges that don't have cells.
            try:
              new_rows[row_index][col_index] = anchor_cell.copy()
            except IndexError:
              pass
    return new_rows

  def rows(self, cell_mode=CellMode.cooked):
    """Generates a sequence of parsed rows from the worksheet. The cells are parsed according to the
    cell_mode argument.
    """
    for row_index in range(self.nrows):
      yield self.parse_row(self.get_row(row_index), row_index, cell_mode)

  def parse_row(self, row, row_index, cell_mode=CellMode.cooked):
    """Parse a row according to the given cell_mode."""
    return [self.parse_cell(cell, (col_index, row_index), cell_mode) \
            for col_index, cell in enumerate(row)]

  def get_cell(self, coords, cell_mode=CellMode.cooked):
    """Get a cell from the worksheet according to the given coords and parse it according to the
    given cell_mode. Coords is a tuple of (col, row).
    """
    return self.parse_cell(self.get_row(coords[1])[coords[0]], coords, cell_mode)

  def tuple_to_datetime(self, date_tuple):
    """Converts a tuple to a either a date, time or datetime object.

    If the Y, M and D parts of the tuple are all 0, then a time object is returned.
    If the h, m and s aprts of the tuple are all 0, then a date object is returned.
    Otherwise a datetime object is returned.
    """
    year, month, day, hour, minute, second = date_tuple
    if year == month == day == 0:
      return time(hour, minute, second)
    elif hour == minute == second == 0:
      return date(year, month, day)
    else:
      return datetime(year, month, day, hour, minute, second)

  def parse_cell(self, cell, coords, mode=CellMode.cooked):
    """Virtual method for parsing a cell. Must be overridden by subclasses."""
    raise NotImplementedError()

  def get_row(self, row_index):
    """Virtual method for getting a row, by index, from a worksheet. Must be overridden by
    subclasses.
    """
    raise NotImplementedError()

  def merged_cell_ranges(self):
    """Enumerates cell ranges as ((cell_low, row_low), (cell_high, row_high)). Can be overridden by
    subclasses."""
    return []

  def get_note(self, coords):
    """Coords are (col, row). Can be overridden by subclasses."""
    return None


class WorkbookBase(object):
  """The base class for all workbooks. For each implemented spreadsheet type, it is necessary to
  write a subclass of this class that implements the necessary methods."""
  def __init__(self, filename):
    """Initialize the WorkbookBase instance."""
    self.filename = filename
    self._sheets = None

  def sheets(self, index=None):
    """Return either a list of all sheets if index is None, or the sheet at the given index."""
    if self._sheets is None:
      self._sheets = [self.get_worksheet(s, i) for i, s in enumerate(self.iterate_sheets())]
    if index is None:
      return self._sheets
    else:
      return self._sheets[index]

  def iterate_sheets(self):
    """Virtual method to iterate through sheets. Must be overridden by child classes."""
    raise NotImplementedError()

  def get_worksheet(self, raw_sheet, index):
    """Virtual method to get a WorksheetBase class from a raw sheet. Must be overridden by child
    classes."""
    return WorksheetBase(raw_sheet, index)



from .test_validation import DataValidationTestCase
from .test_model_base import DataModelBaseTestCase
from .test_sql_writer import DataSQLWriterTestCase
from .test_data_access import DataDataAccessTestCase
from .test_model_access import DataModelAccessTestCase
from .test_foreign_model_wrapper import DataForeignModelWrapperTestCase



from setuptools import setup, find_packages

version = "0.1.11"

setup(name="bidon",
      version=version,
      license="BSD",
      platforms="any",
      author="Trey Cucco",
      author_email="fcucco@gmail.com",
      packages=find_packages(),
      description="A simple, easy to use, and flexible data handling library",
      url="https://github.com/treycucco/bidon",
      download_url="https://github.com/treycucco/bidon/tarball/master",
      classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3 :: Only"
      ])



import unittest

from bidon.data import ModelBase, ModelAccess

from tests import get_model_access


class Person(ModelBase):
  table_name = "people"
  attrs = {
    "first_name": None,
    "last_name": None,
    "age": 0 }


class PersonThing(ModelBase):
  table_name = "peoples_things"
  primary_key_name = ["person_id", "thing_id"]
  primary_key_is_auto = False
  attrs = {
    "person_id": None,
    "thing_id": None,
    "quantity": 0}


class DataModelAccessTestCase(unittest.TestCase):
  @classmethod
  def setUpClass(cls):
    cls.mab = get_model_access()
    cls.mab.open()

  @classmethod
  def tearDownClass(cls):
    cls.mab.rollback()
    cls.mab.close()

  def test_find_model(self):
    p = self.mab.find_model(Person, dict(first_name="Yours"))
    self.assertIsNotNone(p)
    self.assertIsInstance(p, Person)
    self.assertEqual(p.first_name, "Yours")

  def test_find_all_models(self):
    ps = list(self.mab.find_all_models(Person, "last_name like 'T%'"))
    self.assertEqual(len(ps), 2)
    self.assertIsInstance(ps[0], Person)
    self.assertEqual(ps[0].last_name[0], "T")

  def test_page_models(self):
    ps, count = self.mab.page_models(Person, (0, 4), order_by="id")
    self.assertEqual(count, 11)
    self.assertEqual(len(ps), 4)
    self.assertEqual([1, 2, 3, 4], [p.id for p in ps])

  def test_find_model_by_id(self):
    pt = self.mab.find_model_by_id(PersonThing, [1, 1])
    self.assertEqual(pt.person_id, 1)
    self.assertEqual(pt.thing_id, 1)
    self.assertEqual(pt.quantity, 1)
    p = self.mab.find_model_by_id(Person, 1)

  def test_refresh_model(self):
    with get_model_access() as mab:
      pt = mab.find_model(Person, "1=1")
      nn = "aabbccdd"
      mab.update(Person.table_name, dict(first_name=nn), dict(id=pt.id))
      self.assertNotEqual(pt.first_name, nn)
      self.assertEqual(mab.refresh_model(pt, False).first_name, nn)
      self.assertNotEqual(pt.first_name, nn)
      mab.refresh_model(pt, True)
      self.assertEqual(pt.first_name, nn)
      mab.rollback()

  def test_update_model(self):
    with get_model_access() as mab:
      p1 = mab.find_model_by_id(Person, 1)
      ua = p1.updated_at
      self.assertNotEqual(p1.first_name, "Minerva")
      updatedattrs = p1.update(dict(first_name="Minerva", last_name="McGonagall"))
      mab.update_model(p1, include_keys=updatedattrs)
      if mab.core.supports_returning_syntax:
        self.assertNotEqual(ua, p1.updated_at)
      p2 = mab.find_model_by_id(Person, 1)
      self.assertEqual(p2.first_name, "Minerva")
      self.assertEqual(p2.last_name, "McGonagall")
      mab.rollback()

  def test_insert_model(self):
    with get_model_access() as mab:
      p1 = Person(dict(first_name="Albus", last_name="Dumbledore", age=115))
      self.assertIsNone(p1.id)
      mab.insert_model(p1)
      self.assertIsNotNone(p1.id)
      self.assertIsNotNone(p1.created_at)
      p2 = mab.find_model_by_id(Person, p1.id)
      self.assertEqual(p1.first_name, p2.first_name)
      self.assertEqual(p1.last_name, p2.last_name)
      self.assertEqual(p1.age, p2.age)
      mab.rollback()

  def test_delete_model(self):
    with get_model_access() as mab:
      pc = mab.count("people")
      mab.delete_model(Person(id=11))
      self.assertEqual(pc - 1, mab.count("people"))
      self.assertIsNone(mab.find_model_by_id(Person, 11))

      ptc = mab.count("peoples_things")
      mab.delete_model(PersonThing(person_id=1, thing_id=1))
      self.assertEqual(ptc - 1, mab.count("peoples_things"))
      self.assertIsNone(mab.find_model_by_id(PersonThing, [1, 1]))

      mab.delete_model(PersonThing, [1, 2])
      self.assertEqual(ptc - 2, mab.count("peoples_things"))
      self.assertIsNone(mab.find_model_by_id(PersonThing, [1, 1]))

      mab.rollback()
